from __future__ import annotations

import asyncio
import functools
import inspect
from collections import defaultdict
from operator import getitem
from typing import Any, Callable, Dict, List, Optional, Set, TypeVar

import numpy

import dask
import dask.dataframe
from dask.base import tokenize
from dask.dataframe.core import DataFrame, Series, _concat as dd_concat, new_dd_object
from dask.dataframe.shuffle import group_split_dispatch, hash_object_dispatch
from distributed import wait
from distributed.protocol import nested_deserialize, to_serialize
from distributed.worker import Worker

from .. import comms

T = TypeVar("T")


async def send(
    eps,
    myrank,
    rank_to_out_part_ids: Dict[int, Set[int]],
    out_part_id_to_dataframe: Dict[int, DataFrame],
) -> None:
    """Notice, items sent are removed from `out_part_id_to_dataframe`"""
    futures = []
    for rank, out_part_ids in rank_to_out_part_ids.items():
        if rank != myrank:
            msg = {
                i: to_serialize(out_part_id_to_dataframe.pop(i))
                for i in (out_part_ids & out_part_id_to_dataframe.keys())
            }
            futures.append(eps[rank].write(msg))
    await asyncio.gather(*futures)


async def recv(
    eps,
    myrank,
    rank_to_out_part_ids: Dict[int, Set[int]],
    out_part_id_to_dataframe_list: Dict[int, List[DataFrame]],
    proxify,
) -> None:
    """Notice, received items are appended to `out_parts_list`"""

    async def read_msg(rank: int) -> None:
        msg: Dict[int, DataFrame] = nested_deserialize(await eps[rank].read())
        for out_part_id, df in msg.items():
            out_part_id_to_dataframe_list[out_part_id].append(proxify(df))

    await asyncio.gather(
        *(read_msg(rank) for rank in rank_to_out_part_ids if rank != myrank)
    )


def get_proxify(worker: Worker) -> Callable[[T], T]:
    """Get function to proxify objects"""
    from dask_cuda.proxify_host_file import ProxifyHostFile

    if isinstance(worker.data, ProxifyHostFile):
        data = worker.data
        return lambda x: data.manager.proxify(x)[0]
    return lambda x: x  # no-op


def compute_map_index(df: Any, column_names, npartitions) -> Series:
    """Return a Series that maps each row `df` to a partition ID

    The partitions are determined by hashing the columns given by column_names
    unless if `column_names[0] == "_partitions"`, in which case the values of
    `column_names[0]` are used as index.

    Parameters
    ----------
    df: DataFrame
    column_names: list of strings
        List of column names on which we want to split.
    npartitions: int or None
        The desired number of output partitions.

    Returns
    -------
    out: Dict[int, DataFrame]
        A dictionary mapping integers in {0..k} to dataframes such that the
        hash values of `df[col]` are well partitioned.
    """

    if column_names[0] == "_partitions":
        ind = df[column_names[0]]
    else:
        ind = hash_object_dispatch(
            df[column_names] if column_names else df, index=False
        )
    typ = numpy.min_scalar_type(npartitions * 2)
    return (ind % npartitions).astype(typ, copy=False)


def single_shuffle_group(
    df: DataFrame, column_names, npartitions, ignore_index
) -> Dict[int, DataFrame]:
    """Split dataframe based on the indexes returned by `compute_map_index`"""
    map_index = compute_map_index(df, column_names, npartitions)
    return group_split_dispatch(df, map_index, npartitions, ignore_index=ignore_index)


def multi_shuffle_group(
    df_meta: DataFrame,
    dfs: Dict[str, DataFrame],
    column_names,
    npartitions,
    ignore_index,
    proxify,
) -> Dict[int, DataFrame]:
    """Split multiple dataframes such that each partition hashes to the same

    Since we concatenate dataframes belonging to the same partition, each
    partition ID maps to exactly one dataframe.

    Parameters
    ----------
    df_meta: DataFrame
        An empty dataframe matching the expected output
    dfs: dict of dataframes
        The dataframes to split given as a map of stage keys to dataframes
    column_names: list of strings
        List of column names on which we want to split.
    npartitions: int or None
        The desired number of output partitions.
    ignore_index: bool
        Ignore index during shuffle.  If True, performance may improve,
        but index values will not be preserved.
    proxify: callable
        Function to proxify object.

    Returns
    -------
    dict of DataFrames
        Mapping from partition ID to dataframe.
    """

    # Grouping each input dataframe, one part for each partition ID.
    dfs_grouped: List[Dict[int, DataFrame]] = []
    while dfs:
        dfs_grouped.append(
            proxify(
                single_shuffle_group(
                    # pop dataframe in any order, to free staged memory ASAP
                    dfs.popitem()[1],
                    column_names,
                    npartitions,
                    ignore_index,
                )
            )
        )

    # Maps each output partition ID to a dataframe. If the partition is empty,
    # an empty dataframe is used.
    ret: Dict[int, DataFrame] = {}
    for i in range(npartitions):  # Iterate over all possible output partition IDs
        t = [df_grouped[i] for df_grouped in dfs_grouped]
        if len(t) == 1:
            ret[i] = t[0]
        elif len(t) > 1:
            ret[i] = proxify(dd_concat(t, ignore_index=ignore_index))
        else:
            ret[i] = df_meta  # Empty dataframe
    return ret


async def shuffle_task(
    s,
    stage_name,
    df_meta,
    rank_to_inkeys: Dict[int, set],
    rank_to_out_part_ids: Dict[int, Set[int]],
    column_names,
    npartitions,
    ignore_index,
) -> List[DataFrame]:
    """Explicit-comms shuffle task

    This function is running on each worker participating in the shuffle.

    Parameters
    ----------
    s: dict
        Worker session state
    stage_name: str
        Name of the stage to retrieve the input keys from.
    rank_to_inkeys: dict
        dict that for each worker rank specifices the set of staged input keys.
    rank_to_out_part_ids: dict
        dict that for each worker rank specifices a set of output partition IDs.
        If the worker shouldn't return any partitions, it is excluded from the
        dict. Partition IDs are global integers `0..npartitions` and corresponds
        to the dict keys returned by `group_split_dispatch`.
    column_names: list of strings
        List of column names on which we want to split.
    npartitions: int or None
        The desired number of output partitions.
    ignore_index: bool
        Ignore index during shuffle.  If True, performance may improve,
        but index values will not be preserved.

    Returns
    -------
    partitions: list of DataFrames
        List of dataframe-partitions
    """

    proxify = get_proxify(s["worker"])
    myrank = s["rank"]
    eps = s["eps"]
    stage = comms.pop_staging_area(s, stage_name)
    assert stage.keys() == rank_to_inkeys[myrank]

    out_part_id_to_dataframe = multi_shuffle_group(
        df_meta=df_meta,
        dfs=stage,
        column_names=column_names,
        npartitions=npartitions,
        ignore_index=ignore_index,
        proxify=proxify,
    )

    # Communicate all the dataframe-partitions all-to-all. The result is
    # `out_part_id_to_dataframe_list` that for each output partition maps
    # a list of dataframes received.
    out_part_id_to_dataframe_list: Dict[int, List[DataFrame]] = defaultdict(list)
    await asyncio.gather(
        recv(eps, myrank, rank_to_out_part_ids, out_part_id_to_dataframe_list, proxify),
        send(eps, myrank, rank_to_out_part_ids, out_part_id_to_dataframe),
    )

    # At this point `send()` should have pop'ed all output partitions
    # beside the partitions owned be `myrank`.
    assert rank_to_out_part_ids[myrank] == out_part_id_to_dataframe.keys()
    # We can now add them to the output dataframes.
    for out_part_id, dataframe in out_part_id_to_dataframe.items():
        out_part_id_to_dataframe_list[out_part_id].append(dataframe)
    del out_part_id_to_dataframe

    # Finally, we concatenate the output dataframes into the final output partitions
    return [
        proxify(dd_concat(dfs, ignore_index=ignore_index))
        for dfs in out_part_id_to_dataframe_list.values()
    ]


def shuffle(
    df: DataFrame,
    column_names: List[str],
    npartitions: Optional[int] = None,
    ignore_index: bool = False,
) -> DataFrame:
    """Order divisions of DataFrame so that all values within column(s) align

    This enacts a task-based shuffle using explicit-comms. It requires a full
    dataset read, serialization and shuffle. This is expensive. If possible
    you should avoid shuffles.

    This does not preserve a meaningful index/partitioning scheme. This is not
    deterministic if done in parallel.

    Requires an activate client.

    Parameters
    ----------
    df: dask.dataframe.DataFrame
        Dataframe to shuffle
    column_names: list of strings
        List of column names on which we want to split.
    npartitions: int or None
        The desired number of output partitions. If None, the number of output
        partitions equals `df.npartitions`
    ignore_index: bool
        Ignore index during shuffle. If True, performance may improve,
        but index values will not be preserved.

    Returns
    -------
    df: dask.dataframe.DataFrame
        Shuffled dataframe

    Developer Notes
    ---------------
    The implementation consist of three steps:
      (a) Stage the partitions of `df` on all workers and then cancel them
          thus at this point the Dask Scheduler doesn't know about any of the
          the partitions.
      (b) Submit a task on each worker that shuffle (all-to-all communicate)
          the staged partitions and return a list of dataframe-partitions.
      (c) Submit a dask graph that extract (using `getitem()`) individual
          dataframe-partitions from (b).
    """
    c = comms.default_comms()

    # The ranks of the output workers
    ranks = list(range(len(c.worker_addresses)))

    # By default, we preserve number of partitions
    if npartitions is None:
        npartitions = df.npartitions

    # Step (a):
    df = df.persist()  # Make sure optimizations are apply on the existing graph
    wait(df)  # Make sure all keys has been materialized on workers
    name = (
        "explicit-comms-shuffle-"
        f"{tokenize(df, column_names, npartitions, ignore_index)}"
    )
    df_meta: DataFrame = df._meta

    # Stage all keys of `df` on the workers and cancel them, which makes it possible
    # for the shuffle to free memory as the partitions of `df` are consumed.
    # See CommsContext.stage_keys() for a description of staging.
    rank_to_inkeys = c.stage_keys(name=name, keys=df.__dask_keys__())
    c.client.cancel(df)

    # Find the output partition IDs for each worker
    div = npartitions // len(ranks)
    rank_to_out_part_ids: Dict[int, Set[int]] = {}  # rank -> set of partition id
    for i, rank in enumerate(ranks):
        rank_to_out_part_ids[rank] = set(range(div * i, div * (i + 1)))
    for rank, i in zip(ranks, range(div * len(ranks), npartitions)):
        rank_to_out_part_ids[rank].add(i)

    # Run `_shuffle()` on each worker
    shuffle_result = {}
    for rank in ranks:
        shuffle_result[rank] = c.submit(
            c.worker_addresses[rank],
            shuffle_task,
            name,
            df_meta,
            rank_to_inkeys,
            rank_to_out_part_ids,
            column_names,
            npartitions,
            ignore_index,
        )
    wait(list(shuffle_result.values()))

    # Step (d): extract individual dataframe-partitions. We use `submit()`
    #           to control where the tasks are executed.
    # TODO: can we do this without using `submit()` to avoid the overhead
    #       of creating a Future for each dataframe partition?

    dsk = {}
    for rank in ranks:
        for i, part_id in enumerate(rank_to_out_part_ids[rank]):
            dsk[(name, part_id)] = c.client.submit(
                getitem, shuffle_result[rank], i, workers=[c.worker_addresses[rank]]
            )

    # Create a distributed Dataframe from all the pieces
    divs = [None] * (len(dsk) + 1)
    ret = new_dd_object(dsk, name, df_meta, divs).persist()
    wait(ret)

    # Release all temporary dataframes
    for fut in [*shuffle_result.values(), *dsk.values()]:
        fut.release()
    return ret


def get_rearrange_by_column_tasks_wrapper(func):
    """Returns a function wrapper that dispatch the shuffle to explicit-comms.

    Notice, this is monkey patched into Dask at dask_cuda import
    """

    func_sig = inspect.signature(func)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if dask.config.get("explicit-comms", False):
            try:
                import distributed.worker

                # Make sure we have an activate client.
                distributed.worker.get_client()
            except (ImportError, ValueError):
                pass
            else:
                # Convert `*args, **kwargs` to a dict of `keyword -> values`
                kw = func_sig.bind(*args, **kwargs)
                kw.apply_defaults()
                kw = kw.arguments
                column = kw["column"]
                if isinstance(column, str):
                    column = [column]
                return shuffle(kw["df"], column, kw["npartitions"], kw["ignore_index"])
        return func(*args, **kwargs)

    return wrapper



import logging
import os

import click
import numba.cuda

import dask
import distributed.comm.ucx
from distributed.diagnostics.nvml import get_device_index_and_uuid, has_cuda_context

from .utils import get_ucx_config

logger = logging.getLogger(__name__)


def _create_cuda_context_handler():
    if int(os.environ.get("DASK_CUDA_TEST_SINGLE_GPU", "0")) != 0:
        try:
            numba.cuda.current_context()
        except numba.cuda.cudadrv.error.CudaSupportError:
            pass
    else:
        numba.cuda.current_context()


def _create_cuda_context():
    try:
        # Added here to ensure the parent `LocalCUDACluster` process creates the CUDA
        # context directly from the UCX module, thus avoiding a similar warning there.
        try:
            distributed.comm.ucx.init_once()
        except ModuleNotFoundError:
            # UCX intialization has to be delegated to Distributed, it will take care
            # of setting correct environment variables and importing `ucp` after that.
            # Therefore if ``import ucp`` fails we can just continue here.
            pass

        cuda_visible_device = get_device_index_and_uuid(
            os.environ.get("CUDA_VISIBLE_DEVICES", "0").split(",")[0]
        )
        ctx = has_cuda_context()
        if (
            ctx.has_context
            and not distributed.comm.ucx.cuda_context_created.has_context
        ):
            distributed.comm.ucx._warn_existing_cuda_context(ctx, os.getpid())

        _create_cuda_context_handler()

        if not distributed.comm.ucx.cuda_context_created.has_context:
            ctx = has_cuda_context()
            if ctx.has_context and ctx.device_info != cuda_visible_device:
                distributed.comm.ucx._warn_cuda_context_wrong_device(
                    cuda_visible_device, ctx.device_info, os.getpid()
                )

    except Exception:
        logger.error("Unable to start CUDA Context", exc_info=True)


def initialize(
    create_cuda_context=True,
    enable_tcp_over_ucx=None,
    enable_infiniband=None,
    enable_nvlink=None,
    enable_rdmacm=None,
):
    """Create CUDA context and initialize UCX-Py, depending on user parameters.

    Sometimes it is convenient to initialize the CUDA context, particularly before
    starting up Dask worker processes which create a variety of threads.

    To ensure UCX works correctly, it is important to ensure it is initialized with the
    correct options. This is especially important for the client, which cannot be
    configured to use UCX with arguments like ``LocalCUDACluster`` and
    ``dask-cuda-worker``. This function will ensure that they are provided a UCX
    configuration based on the flags and options passed by the user.

    This function can also be used within a worker preload script for UCX configuration
    of mainline Dask.distributed.
    https://docs.dask.org/en/latest/setup/custom-startup.html

    You can add it to your global config with the following YAML:

    .. code-block:: yaml

        distributed:
          worker:
            preload:
              - dask_cuda.initialize

    See https://docs.dask.org/en/latest/configuration.html for more information about
    Dask configuration.

    Parameters
    ----------
    create_cuda_context : bool, default True
        Create CUDA context on initialization.
    enable_tcp_over_ucx : bool, default None
        Set environment variables to enable TCP over UCX, even if InfiniBand and NVLink
        are not supported or disabled.
    enable_infiniband : bool, default None
        Set environment variables to enable UCX over InfiniBand, implies
        ``enable_tcp_over_ucx=True`` when ``True``.
    enable_nvlink : bool, default None
        Set environment variables to enable UCX over NVLink, implies
        ``enable_tcp_over_ucx=True`` when ``True``.
    enable_rdmacm : bool, default None
        Set environment variables to enable UCX RDMA connection manager support,
        requires ``enable_infiniband=True``.
    """
    ucx_config = get_ucx_config(
        enable_tcp_over_ucx=enable_tcp_over_ucx,
        enable_infiniband=enable_infiniband,
        enable_nvlink=enable_nvlink,
        enable_rdmacm=enable_rdmacm,
    )
    dask.config.set({"distributed.comm.ucx": ucx_config})

    if create_cuda_context:
        _create_cuda_context()


@click.command()
@click.option(
    "--create-cuda-context/--no-create-cuda-context",
    default=False,
    help="Create CUDA context",
)
@click.option(
    "--enable-tcp-over-ucx/--disable-tcp-over-ucx",
    default=False,
    help="Enable TCP communication over UCX",
)
@click.option(
    "--enable-infiniband/--disable-infiniband",
    default=False,
    help="Enable InfiniBand communication",
)
@click.option(
    "--enable-nvlink/--disable-nvlink",
    default=False,
    help="Enable NVLink communication",
)
@click.option(
    "--enable-rdmacm/--disable-rdmacm",
    default=False,
    help="Enable RDMA connection manager, currently requires InfiniBand enabled.",
)
def dask_setup(
    service,
    create_cuda_context,
    enable_tcp_over_ucx,
    enable_infiniband,
    enable_nvlink,
    enable_rdmacm,
):
    if create_cuda_context:
        _create_cuda_context()



import pytest

from distributed.sizeof import safe_sizeof

from dask_cuda.device_host_file import DeviceHostFile
from dask_cuda.is_spillable_object import is_spillable_object
from dask_cuda.proxify_host_file import ProxifyHostFile

cupy = pytest.importorskip("cupy")
pandas = pytest.importorskip("pandas")

pytest.importorskip(
    "cudf.core.buffer.spill_manager",
    reason="Current version of cudf doesn't support built-in spilling",
)

import cudf  # noqa: E402
from cudf.core.buffer.spill_manager import (  # noqa: E402
    SpillManager,
    get_global_manager,
    set_global_manager,
)
from cudf.testing._utils import assert_eq  # noqa: E402

if get_global_manager() is not None:
    pytest.skip(
        reason=(
            "cannot test cudf built-in spilling, if already enabled globally. "
            "Please set the `CUDF_SPILL=off` environment variable."
        ),
        allow_module_level=True,
    )


@pytest.fixture
def manager(request):
    """Fixture to enable and make a spilling manager availabe"""
    kwargs = dict(getattr(request, "param", {}))
    set_global_manager(manager=SpillManager(**kwargs))
    yield get_global_manager()
    set_global_manager(manager=None)


def test_is_spillable_object_when_cudf_spilling_disabled():
    pdf = pandas.DataFrame({"a": [1, 2, 3]})
    cdf = cudf.DataFrame({"a": [1, 2, 3]})
    assert not is_spillable_object(pdf)
    assert not is_spillable_object(cdf)


def test_is_spillable_object_when_cudf_spilling_enabled(manager):
    pdf = pandas.DataFrame({"a": [1, 2, 3]})
    cdf = cudf.DataFrame({"a": [1, 2, 3]})
    assert not is_spillable_object(pdf)
    assert is_spillable_object(cdf)


def test_device_host_file_when_cudf_spilling_is_disabled(tmp_path):
    tmpdir = tmp_path / "storage"
    tmpdir.mkdir()
    dhf = DeviceHostFile(
        device_memory_limit=1024 * 16,
        memory_limit=1024 * 16,
        worker_local_directory=tmpdir,
    )
    dhf["pandas"] = pandas.DataFrame({"a": [1, 2, 3]})
    dhf["cudf"] = cudf.DataFrame({"a": [1, 2, 3]})

    assert set(dhf.others.keys()) == set()
    assert set(dhf.device.keys()) == set(["cudf"])
    assert set(dhf.host.keys()) == set(["pandas"])
    assert set(dhf.disk.keys()) == set()


def test_device_host_file_step_by_step(tmp_path, manager: SpillManager):
    tmpdir = tmp_path / "storage"
    tmpdir.mkdir()
    pdf = pandas.DataFrame({"a": [1, 2, 3]})
    cdf = cudf.DataFrame({"a": [1, 2, 3]})
    dhf = DeviceHostFile(
        device_memory_limit=safe_sizeof(pdf),
        memory_limit=safe_sizeof(pdf),
        worker_local_directory=tmpdir,
    )
    dhf["pa1"] = pdf
    dhf["cu1"] = cdf

    assert set(dhf.others.keys()) == set(["cu1"])
    assert set(dhf.device.keys()) == set()
    assert set(dhf.host.keys()) == set(["pa1"])
    assert set(dhf.disk.keys()) == set()
    assert_eq(dhf["pa1"], dhf["cu1"])

    dhf["pa2"] = pdf
    assert set(dhf.others.keys()) == set(["cu1"])
    assert set(dhf.device.keys()) == set()
    assert set(dhf.host.keys()) == set(["pa2"])
    assert set(dhf.disk.keys()) == set(["pa1"])

    dhf["cu2"] = cdf
    assert set(dhf.others.keys()) == set(["cu1", "cu2"])
    assert set(dhf.device.keys()) == set()
    assert set(dhf.host.keys()) == set(["pa2"])
    assert set(dhf.disk.keys()) == set(["pa1"])

    del dhf["cu1"]
    assert set(dhf.others.keys()) == set(["cu2"])
    assert set(dhf.device.keys()) == set()
    assert set(dhf.host.keys()) == set(["pa2"])
    assert set(dhf.disk.keys()) == set(["pa1"])

    del dhf["pa2"]
    assert set(dhf.others.keys()) == set(["cu2"])
    assert set(dhf.device.keys()) == set()
    assert set(dhf.host.keys()) == set()
    assert set(dhf.disk.keys()) == set(["pa1"])

    del dhf["pa1"]
    assert set(dhf.others.keys()) == set(["cu2"])
    assert set(dhf.device.keys()) == set()
    assert set(dhf.host.keys()) == set()
    assert set(dhf.disk.keys()) == set()

    del dhf["cu2"]
    assert set(dhf.others.keys()) == set()
    assert set(dhf.device.keys()) == set()
    assert set(dhf.host.keys()) == set()
    assert set(dhf.disk.keys()) == set()


def test_proxify_host_file(tmp_path_factory, manager: SpillManager):
    # Reuse the spill-to-disk dir, if it exist
    if ProxifyHostFile._spill_to_disk is None:
        tmpdir = tmp_path_factory.mktemp("jit-unspill")
    else:
        tmpdir = ProxifyHostFile._spill_to_disk.root_dir / ".."

    with pytest.warns(
        UserWarning,
        match="JIT-Unspill and cuDF's built-in spilling don't work together",
    ):
        dhf = ProxifyHostFile(
            device_memory_limit=1000,
            memory_limit=1000,
            worker_local_directory=str(tmpdir),
        )
    dhf["cu1"] = cudf.DataFrame({"a": [1, 2, 3]})
    del dhf["cu1"]
